#include <stdio.h>
#include <stdlib.h>

#include "../turing/turing.h"
#include "tokens.h"

#define POMAK_DESNO	'R'
#define POMAK_LIJEVO	'L'
#define POMAK_STOJI	'S'

const TuringUlaz UVIJEK = TuringUlaz();

static void IzbaciGresku(const string & greska) {
	fprintf(stderr, "%s\n", greska.c_str());
	exit(1);
}

class Ctx {
	public:
		typedef pair<const TuringStanje **, string> GoToPar;

		CTokIter tok;
		int cs;
		map<const string, TuringStanje> stanje;
		unsigned int numtraka;
		string NOPOMAK;

		//const TuringStanje *prev_begin;
		Ctx() {
			cs = 0;
			numtraka = 0;
			prvostanje = NULL;
		}

		void PristupTraci(int traka, string * koristeni = NULL) {
			if ((traka + 1) <= 0) {
				IzbaciGresku("Nevaljan broj trake!");
			}

			if (numtraka >= (unsigned int)(traka + 1)) return;

			numtraka = traka + 1;
			NOPOMAK = string(numtraka, POMAK_STOJI);

			if (koristeni) {
				*koristeni += string(numtraka - koristeni->size(), POMAK_STOJI);
			}
		}

		void AddGoTo(const TuringStanje **tsp, const string & naziv) {
			gotorefs.push_back(GoToPar(tsp, naziv));
		}

		void UpdateGoToRefs() {
			for (vector<GoToPar>::const_iterator i = gotorefs.begin(); i != gotorefs.end(); ++i) {
				map<const string, TuringStanje>::const_iterator st;
				if ((st = stanje.find(i->second)) == stanje.end()) {
					IzbaciGresku("Referencirano stanje \"" + i->second + "\" ne postoji!");
				}
				*i->first = &st->second;
			}
		}

		TuringStanje & NovoStanje(const char * naziv = NULL) {
			string naz;
			if (naziv) {
				naz = naziv;
				if (naz.empty()) {
					IzbaciGresku("Ne moze se postaviti prazno ime stanju!");
				}
				if (stanje.find(naz) != stanje.end()) {
					IzbaciGresku("Stanje s imenom \"%s\" vec postoji!" + naz);
				}
			} else {
				char cnaz[100];
				sprintf(cnaz, "q%d", cs++);
				naz = cnaz;
			}
			TuringStanje & ts = stanje[naz];
			ts.naziv = &stanje.find(naz)->first;
			ts.definirano = true;
			if (!prvostanje) prvostanje = &ts;
			return ts;
		}

	private:
		vector<GoToPar> gotorefs;
		const TuringStanje *prvostanje;

	friend void OutputResult(FILE *f, const Ctx &C);
};

class ProcRes {
	public:
		const TuringStanje *poc;
		vector<const TuringStanje **>kraj;

		void SetKraj(const TuringStanje *s) {
			for (vector<const TuringStanje**>::iterator i = kraj.begin(); i != kraj.end(); ++i) {
				**i = s;
			}
		}

		void DodajKrajeve(const vector<const TuringStanje **> &dk) {
			for (vector<const TuringStanje **>::const_iterator i = dk.begin(); i != dk.end(); ++i)
				kraj.push_back(*i);
		}
};

static char DekodPomak(char c) {
	switch (c) {
		case '>': return POMAK_DESNO;
		case '<': return POMAK_LIJEVO;
		case '.': return POMAK_STOJI;
	}
	IzbaciGresku(string("Koristen je nepostojeci pomak: \'") + c + "\'!");
	return POMAK_STOJI;
}

static ProcRes ProcesirajListu(Ctx &C);

static ProcRes Procesiraj(Ctx &C) {
	ProcRes r;
	const char *naziv_stanja = NULL;

	if (*C.tok->c_str() == ':') {
		naziv_stanja = C.tok->c_str() + 1;
		++C.tok;
	}

	TuringStanje &s = C.NovoStanje(naziv_stanja);
	r.poc = &s;

	switch (*C.tok->c_str()) {
		// N* N{ traka znak smjer }
		case '*': {
			TuringAkcija a;
			a.pomak = C.NOPOMAK;
			for (size_t i = 0, n = C.tok->size(); i < n; ++i) {
				int t;
				t = (++C.tok)->ToInt();
				C.PristupTraci(t, &a.pomak);
				a.zapis[t] = *(++C.tok)->c_str();
				a.pomak[t] = DekodPomak(*(++C.tok)->c_str());
			}
			++C.tok;
			s.akcija[UVIJEK] = a;
			r.kraj.push_back(&s.akcija[UVIJEK].iduce);
			break;
		}

		// N[<>.] N{ traka }
		case '.':
		case '>':
		case '<': {
			TuringAkcija a;
			a.pomak = C.NOPOMAK;
			string smjerovi = *C.tok;
			++C.tok;
			for (size_t i = 0; i < smjerovi.size(); ++i) {
				int t = (C.tok++)->ToInt();
				C.PristupTraci(t, &a.pomak);
				a.pomak[t] = DekodPomak(smjerovi[i]);
			}
			s.akcija[UVIJEK] = a;
			r.kraj.push_back(&s.akcija[UVIJEK].iduce);
			break;
		}

		// [ traka smjer prihvacani @... @... ]
		case '[': {
			++C.tok;
			int traka = C.tok->ToInt();
			C.PristupTraci(traka);
			++C.tok;
			char smjer = DekodPomak(*C.tok->c_str());
			++C.tok;
			for (const char *t = C.tok->c_str(); *t; ++t) {
				TuringAkcija a;
				TuringUlaz u;
				a.pomak = C.NOPOMAK;
				a.pomak[traka] = smjer;
				a.iduce = &s;
				u[traka] = *t;
				s.akcija[u] = a;
			}
			TuringAkcija a;
			a.pomak = C.NOPOMAK;
			s.akcija[UVIJEK] = a;
			r.kraj.push_back(&s.akcija[UVIJEK].iduce);
			++C.tok;

			while (*C.tok == "@") {
				++C.tok;
				const char *ulz = C.tok->c_str();
				++C.tok;
				TuringAkcija a;
				a.pomak = C.NOPOMAK;
				a.pomak[traka] = smjer;
				ProcRes r2 = ProcesirajListu(C);
				r2.SetKraj(&s);
				a.iduce = r2.poc;

				for (; *ulz; ++ulz) {
					TuringUlaz u;
					u[traka] = *ulz;
					s.akcija[u] = a;
				}
			}

			if (*C.tok != "]") {
				IzbaciGresku("Ocekivana ] ili @, a ne " + *C.tok + "!");
			}
			++C.tok;
			break;
		}

		// ( traka pomak @ .. @ .. )
		case '(': {
			++C.tok;
			int t = (C.tok++)->ToInt();
			C.PristupTraci(t);
			char smjer = DekodPomak(*(C.tok++)->c_str());

			TuringAkcija a;
			a.pomak = C.NOPOMAK;
			s.akcija[UVIJEK] = a;
			r.kraj.push_back(&s.akcija[UVIJEK].iduce);

			while (*C.tok == "@") {
				++C.tok;
				const char *ulz = C.tok->c_str();
				++C.tok;
				TuringAkcija a;
				a.pomak = C.NOPOMAK;
				a.pomak[t] = smjer;
				ProcRes r2 = ProcesirajListu(C);
				r.DodajKrajeve(r2.kraj);
				a.iduce = r2.poc;
				for (; *ulz; ++ulz) {
					TuringUlaz u;
					u[t] = *ulz;
					s.akcija[u] = a;
				}
			}

			if (*C.tok != ")") {
				IzbaciGresku("Ocekivana ) ili @, a ne " + *C.tok + "!");
			}
			++C.tok;

			break;
		}

		// Y labela
		case 'Y': {
			++C.tok;
			TuringAkcija a;
			a.pomak = C.NOPOMAK;
			s.akcija[UVIJEK] = a;
			C.AddGoTo(&s.akcija[UVIJEK].iduce, *C.tok);
			++C.tok;
			break;
		}

		default:
			IzbaciGresku("Invalidna operacija: " + *C.tok);
	}
	return r;
}

static ProcRes ProcesirajListu(Ctx &C) {
	ProcRes r, r2, r3;
	r.poc = NULL;
	while (!C.tok.eof() && *C.tok != "@" && *C.tok != "]" && *C.tok != ")") {
		r2 = Procesiraj(C);
		if (!r.poc) {
			r.poc = r2.poc;
		} else {
			r3.SetKraj(r2.poc);
		}
		r3 = r2;
	}
	r.kraj = r2.kraj;
	return r;
}

static void OutputResultStanje(FILE *f, const TuringStanje &st, const Ctx &C) {
	fprintf(f, "%s {\n", st.GetNaziv().c_str());
	for (map<const TuringUlaz, TuringAkcija>::const_iterator j = st.akcija.begin(); j != st.akcija.end(); ++j) {
		fprintf(f, ": ");
		for (unsigned int k = 0; k < C.numtraka; ++k) {
			map<const int, char>::const_iterator l = j->first.find(k);
			if (l != j->first.end())
				fprintf(f, "'%c' ", l->second);
			else
				fprintf(f, "''  ");
		}
		fprintf(f, "\t");
		for (unsigned int k = 0; k < C.numtraka; ++k) {
			map<const int, char>::const_iterator l = j->second.zapis.find(k);
			if (l != j->second.zapis.end())
				fprintf(f, "'%c' ", l->second);
			else
				fprintf(f, "''  ");
		}
		{
			string pomak = j->second.pomak;
			if (C.numtraka > pomak.size())
				pomak += string(C.numtraka - pomak.size(), POMAK_STOJI);
			fprintf(f, "%s ", pomak.c_str());
		}
		if (j->second.iduce)
			fprintf(f, "%s\n", j->second.iduce->GetNaziv().c_str());
		else
			fprintf(f, "kraj\n");
	}
	fprintf(f, "}\n\n");
}

void OutputResult(FILE *f, const Ctx &C) {
	if (!C.prvostanje) {
		IzbaciGresku("Nema nista za generirati!");
		return;
	}

	fprintf(f, "%d tracks\n\n", C.numtraka);

	OutputResultStanje(f, *C.prvostanje, C);
	for (map<const string, TuringStanje>::const_iterator i = C.stanje.begin(); i != C.stanje.end(); ++i) {
		if (&i->second != C.prvostanje)
			OutputResultStanje(f, i->second, C);
	}
}

static void help(const char *exe) {
	printf("Uporaba: %s izvorni_kod\n", exe);
}

int main(int argn, const char *argv[]) {
	Kod K;
	Ctx C;

	if (argn != 2) {
		help(argv[0]);
		return 1;
	}

	if (!K.Load(argv[1])) {
		fprintf(stderr, "File not found!\n");
		return 1;
	}

	C.tok = K.GetIter();

	ProcesirajListu(C);
	C.UpdateGoToRefs();

	OutputResult(stdout, C);

	return 0;
}
