#ifndef _TURING_H_
#define _TURING_H_

#include <stdio.h>
#include <string>
#include <map>

#include <stdlib.h>
#define TuringGreska(s) { \
		fprintf(stderr, "%s\n", s); \
		exit(1); \
		}

using namespace std;

class TuringStanje;
typedef map<const string, TuringStanje> ListaStanja;

class TuringTraka : private string {
	public:
		typedef void (*HandlerFunction)(TuringTraka *, void *);
		class Handler {
			public:
				HandlerFunction f;
				void *data;

				Handler(HandlerFunction _func = NULL, void *_data = NULL) : f(_func), data(_data) {}
		};

		TuringTraka() : trenutno(0U) {}
		char Procitaj() const;
		char operator [] (unsigned int pos) const;
		void Zapisi(char c);
		void Pomak(char c);
		void Pozicioniraj(unsigned int pos);
		unsigned int GetPozicija() const { return trenutno; }
		static int DekodPomak(char c);
		void RegisterChangeHandler(const Handler &h);
		void RegisterMoveHandler(const Handler &h);
		void erase();
		size_t size() const { return string::size(); }
		bool empty() const { return string::empty(); }
		const char * c_str() const { return string::c_str(); }

		void Dump() const;

		string & operator = (const string & s) {
			dynamic_cast<string&>(*this) = s;
			CallHandler(change_handler);
			return *this;
		}

	private:
		Handler change_handler, move_handler;
		unsigned int trenutno;

		void CallHandler(const Handler &h);
};

class TuringAkcija {
	public:
		string pomak;
		map<const int, char> zapis;
		const TuringStanje * iduce;

		TuringAkcija() : iduce(NULL) {}

		void Dump() const;
};

class TuringUlaz : public map<const int, char> {
	public:
		bool Provjeri(const TuringTraka *traka) const;
		void Dump() const;
};

class TuringStanje {
	public:
		typedef map<const TuringUlaz, TuringAkcija> ListaAkcija;

		bool definirano;
		const string *naziv;
		ListaAkcija akcija;

		TuringStanje() : definirano(false), naziv(NULL) {}
		string GetNaziv() const { return naziv ? *naziv : "< ? >"; }
		void Dump() const;
};

class Turing {
	public:
		typedef void (*HandlerFunction)(Turing *stroj, void *data);

		class Handler {
			public:
				HandlerFunction f;
				void *data;

				Handler(HandlerFunction _f = NULL, void *_dat = NULL) : f(_f), data(_dat) {}
		};

		TuringTraka *Traka;

		Turing(unsigned int _numtraka = 1) : trenutno(NULL), pocetno(NULL) {
			Traka = new TuringTraka[numtraka = _numtraka];
		}
		~Turing() {
			delete [] Traka;
		}
		Turing & operator = (const Turing & kopija) {
			TuringGreska("Ne moze se kopirati Turingov stroj!");
			return *this;
		}
		bool Ucitaj(const string & fname);
		bool IsReady() const;
		void Init();
		void Simuliraj();
		int Trace();
		void DumpMemory() const;
		int GetNumTraka() const { return numtraka; }
		const ListaStanja & GetStanja() const { return stanje; }
		const TuringStanje * GetPocetno() const { return pocetno; }
		const TuringStanje * FindStanje(const string &name) const;
		void RegisterTraceHandler(const Handler &h);
		void UnregisterTraceHandler();
		const TuringStanje * SetTrenutno(const TuringStanje *nts);
		const TuringStanje * GetTrenutno() const { return trenutno; }

	private:
		int numtraka;
		ListaStanja stanje;
		const TuringStanje *trenutno;
		const TuringStanje *pocetno;
		Handler trace_handler;

		void CallHandler(const Handler &h);
		TuringAkcija UcitajAkciju(FILE *f);
		TuringUlaz UcitajUlaz(FILE *f) const;
		const TuringStanje * Izvrsi(const TuringStanje *st, int *err = NULL);
		void NoveTrake(int _numtraka) {
			delete [] Traka;
			Traka = new TuringTraka[numtraka = _numtraka];
		}
};

#endif
